/* =====================================================================================
 * OS for ACTOframework TUIO
 * =====================================================================================
 * firmware to provide communication with ACTO_framework via Nordic nRF24L01 chip
 * using Mirf, a Arduino library for communication with nRF24L01
 * http://playground.arduino.cc/InterfacingWithHardware/Nrf24L01
 *
 * see also ACTOControl.h or ACTO_framework for protocol
 *
 * date: 05. Mai 2014
 * version: v1.3
 *
 * project: ACTO
 * author: Emanuel Vonach
 * Interactive Media Systems Group, Vienna University of Technology, Austria
 * www.ims.tuwien.ac.at
 * Contact: vonach@ims.tuwien.ac.at, gerstweiler@ims.tuwien.ac.at
 * =====================================================================================
 */

#include <Mirf.h>
#include <MirfHardwareSpiDriver.h>  //MMI
#include <MirfSpiDriver.h>  //MMI
#include <nRF24L01.h>

#include <SPI.h>

///////////////////////////////////////////////////////////////////////////////////////
// TUIO SETTINGS
///////////////////////////////////////////////////////////////////////////////////////

// for serial communication
const long baudRate = 115200;

//for Mirf communication
const byte mirfHomeAddress[5] = {'t','u','i','o',tuioId};   //own id
const byte mirfAwayAddress[5] = {'b','a','s','e','1'};   //base

// variables for errorHandler
unsigned long errorTimestamp = 0;
int blinkNumber = 0;
int ledState = LOW;

//variables for batteryChecker
unsigned long batteryTimestamp = 0;

// for Mirf
byte inMirfBuffer[mirfPayload];
byte outMirfBuffer[mirfPayload];
int outMirfIndex = 0;

//update rate of digital and analog inputs
const unsigned long updateRate = 50;//ms
unsigned long updateRateTimer = 0;

// for extensions units
const int sizeOfDigitalInput = sizeof(digitalInputPins)/2;//sizeof(typeof(digitalInputPins));
const int sizeOfAnalogInput = sizeof(analogInputPins)/2;//sizeof(typeof(analogInputPins));
const int sizeOfDigitalOutput = sizeof(digitalOutputPins)/2;//sizeof(typeof(digitalOutputPins));

// store data from inputs
int digitalInputData[sizeOfDigitalInput];
int analogInputData[sizeOfAnalogInput];

//initialize data arrays which should be addressable by ACTO_framework
byte _byteData[sizeOfByteData];
char _charData[sizeOfCharData];
int _intData[sizeOfIntData];
float _floatData[sizeOfFloatData];


///////////////////////////////////////////////////////////////////////////////////////
// ACTO_framework
///////////////////////////////////////////////////////////////////////////////////////

void setupActoFramework() {
    
  //set motor idle on startup
  initMotorUnit();

  //setup pins & stuff
  if (STATUSLED >= 0) pinMode(STATUSLED, OUTPUT);
  
  //check if developer setup is ok
  if(!checkDeveloperSetup()) {
    errorHandler(SETUPERROR);     //stop if developer setup is not ok
    while(true) errorHandler(0);  //get stuck and blink error
  }
  
  for (int i=0; i < sizeOfDigitalOutput; i++) pinMode(digitalOutputPins[i], OUTPUT);
  for (int i=0; i < sizeOfDigitalInput; i++) pinMode(digitalInputPins[i], INPUT);
  
  //ONLY DEBUG
  Serial.begin(baudRate);
    
  // MIRF SETUP  
  // ------------------------------------------------------------------------------------------------------------------------------// 
  // nRF24L01 |   ATmega8      | ATmega168      |    ARDUINO Diecimila (ATmega168)  |        Description                           //  
  // ------------------------------------------------------------------------------------------------------------------------------// 
  // SCK      |   PB5 (SCK)    |  PB5 (SCK)     |    digital pin 13                 |        SPI clock signal                      //  
  // MISO     |   PB4 (MISO)   |  PB4 (MISO)    |    digital pin 12                 |        SPI master in, slave out              //
  // MOSI     |   PB3 (MOSI)   |  PB3 (MOSI)    |    digital pin 11                 |        SPI master out, slave in              //
  // CSN      |   PB1          |  PB1           |    digital pin 9                  |        SPI Chip Select                       //
  // CE       |   PB0          |  PB0           |    digital pin 8                  |        Chip Enable Activates RX or TX mode   //
  // IRQ      |   PD2 (INT0)   |  PD6 (PCINT22) |    digital pin 6                  |        Interrupt pin. Active low             //
  // ------------------------------------------------------------------------------------------------------------------------------// 

  // setup pins / SPI
  Mirf.csnPin = 9;
  Mirf.cePin = 8;
  Mirf.spi = &MirfHardwareSpi;
  Mirf.init();

  // Configure Mirf parameters      
  Mirf.setRADDR((byte*)mirfHomeAddress);
  Mirf.setTADDR((byte*)mirfAwayAddress);
  Mirf.payload = mirfPayload;
 
  // Write channel and payload config then power up reciver.
  Mirf.config();
}

boolean checkDeveloperSetup() {  
  //check if output and input pins are separate (could damage board!)
  if (sizeOfDigitalInput > 0 && sizeOfDigitalOutput > 0) {
    for(int i = 0; i < sizeOfDigitalInput; i++) {
      for(int k = 0; k < sizeOfDigitalOutput; k++) {
        if(digitalInputPins[i] == digitalOutputPins[k]) {
          return false;
        }
      }
    }
  }
  //probably check if id is ok?
  return true;
}

void loopActoFramework() {
  // if an error has occured previously, the LED changes state
  errorHandler(0);
  if(checkBattery()) {
    updateActoInputs();
    motorLoop();
    readFromServer();
  }
}

boolean setByteData(byte id, byte val) {
  if (id < sizeOfByteData) {
      writeByteMsg(USB_SET_BYTE, id, val);
      _byteData[id] = val;
      return true;
  }
  else
    return false;
}
byte getByteData(byte id) {
  if (id < sizeOfByteData) {
    return _byteData[id];
  }
  else
    return false;
}
boolean setCharData(byte id, char val) {
  if (id < sizeOfCharData) {
      writeCharMsg(USB_SET_CHAR, id, val);
      _charData[id] = val;
      return true;
  }
  else
    return false;
}
byte getCharData(byte id) {
  if (id < sizeOfCharData) {
    return _charData[id];
  }
  else
    return false;
}
boolean setIntData(byte id, int val) {
  if (id < sizeOfIntData) {
      writeIntMsg(USB_SET_INT, id, val);
      _intData[id] = val;
      return true;
  }
  else
    return false;
}
byte getIntData(byte id) {
  if (id < sizeOfIntData) {
    return _intData[id];
  }
  else
    return false;
}
boolean setFloatData(byte id, float val) {
  if (id < sizeOfFloatData) {
      writeFloatMsg(USB_SET_FLOAT, id, val);
      _floatData[id] = val;
      return true;
  }
  else
    return false;
}
byte getFloatData(byte id) {
  if (id < sizeOfFloatData) {
    return _floatData[id];
  }
  else
    return false;
}

//use this instead of delay(ms)
void actoDelay (int pauseTime) {
  unsigned long startTime = millis();
  while ((millis() - startTime) <= pauseTime) {
    loopActoFramework();
  }
}

//automatically update all inputs defined in digitalInputPins & analogInputPins, in intervall of updateRate
void updateActoInputs() {
  for (int i=0; i < sizeOfDigitalInput; i++){        //use update rate as well?
    int newVal = digitalRead(digitalInputPins[i]);
    if (newVal != digitalInputData[i])
    {
      digitalInputData[i] = newVal;
      onDigitalInputEvent((byte) i, (byte)newVal);
      writeByteMsg(USB_SET_DI_UNIT, (byte)i,(byte)newVal);
    }
  }
  if (millis() - updateRateTimer >= updateRate) {    //minimum update rate of analogPins is 10000 times per second
    for (int i=0; i < sizeOfAnalogInput; i++){
      int newVal = analogRead(analogInputPins[i]);
      if (analogInputData[i] != newVal)              //probably use threshhold of +-1???
      {
        analogInputData[i] = newVal;
        onAnalogInputEvent((byte) i, newVal);
        writeIntMsg(USB_SET_AN_UNIT, (byte)i, newVal);                 //mirfUpdateRate required??? or proper handshake
      }
    }
    updateRateTimer = millis();
  }
}

void readFromServer () {
  if(Mirf.dataReady()){  //there should be no problem with crosstalk, since message is in payload as a whole
    //read message
    Mirf.getData((byte *)&inMirfBuffer);

    switch (inMirfBuffer[0]) {
//      case USB_ACK: case USB_NAK:  // 2- format: CMD, SEP
//        errorHandler(DEBUG);
//        break;
        
      case USB_PING: // 3- format: CMD, ID, SEP
        writeSimpleMsg(USB_PING);
        break;
        
//      case USB_SYNC:  // 3- format: CMD, ID, SEP
//        //do full sync, if necessary
//        errorHandler(DEBUG);
//        break;
//        
      case USB_MOTOR: // 5- format: CMD, ID, direction(byte), speed(byte), SEP
        driveDirectionSpeed(inMirfBuffer[2], inMirfBuffer[3]);
        writeSimpleMsg(USB_PING);
        break;
      
      case USB_GET_DI_UNIT:  //format: 4- CMD, ID, dataID(byte), SEP
        if (inMirfBuffer[2] < sizeOfDigitalInput)
          writeByteMsg(USB_SET_DI_UNIT, inMirfBuffer[2], (byte)digitalInputData[inMirfBuffer[2]]);
        else writeMirfAck(false);
        break;
        
      case USB_GET_AN_UNIT:  //format: 4- CMD, ID, dataID(byte), SEP
        if (inMirfBuffer[2] < sizeOfAnalogInput)
          writeIntMsg(USB_SET_AN_UNIT, inMirfBuffer[2], analogInputData[inMirfBuffer[2]]);
        else writeMirfAck(false);
        break;
      
      case USB_SET_DI_UNIT: // 5- format: CMD, ID, dataID(byte), data(byte), SEP
        
        if (inMirfBuffer[2] < sizeOfDigitalOutput) {
          digitalWrite(digitalOutputPins[(int)inMirfBuffer[2]], (int)inMirfBuffer[3]);
        }
        else writeMirfAck(false);
        break;
        
//      case USB_SET_AN_UNIT: // ONLY ANSWER, should be removed!! 8- format: CMD, ID, dataID(byte), data(int), SEP
//        errorHandler(DEBUG);
//        break;
    
      case USB_SET_PWM_UNIT: // 8- format: CMD, ID, dataID(byte), data(int), SEP
        if (inMirfBuffer[2] < sizeOfDigitalOutput) {
          ActoInt data32i;
          for (int i=0; i < 4; i++){//2
            data32i.binary[i] = inMirfBuffer[3+i];
          }
          analogWrite(digitalOutputPins[inMirfBuffer[2]], data32i.integer);
        }
        else writeMirfAck(false);
        break;
  
      case USB_GET_BYTE:  //format: 4- CMD, ID, dataID(byte), SEP
        if (inMirfBuffer[2] < sizeOfByteData)
          writeByteMsg(USB_SET_BYTE, inMirfBuffer[2], _byteData[inMirfBuffer[2]]);
          
        else writeMirfAck(false);
        break;
        
      case USB_GET_CHAR:  //format: 4- CMD, ID, dataID(byte), SEP
        if (inMirfBuffer[2] < sizeOfCharData)
          writeCharMsg(USB_SET_CHAR, inMirfBuffer[2], _charData[inMirfBuffer[2]]);
        else writeMirfAck(false);
        break;
     
      case USB_GET_INT:  //format: 4- CMD, ID, dataID(byte), SEP
        if (inMirfBuffer[2] < sizeOfIntData)
          writeIntMsg(USB_SET_INT, inMirfBuffer[2], _intData[inMirfBuffer[2]]);
        else writeMirfAck(false);
        break;
        
      case USB_GET_FLOAT:  //format: 4- CMD, ID, dataID(byte), SEP
        if (inMirfBuffer[2] < sizeOfFloatData)
          writeFloatMsg(USB_SET_FLOAT, inMirfBuffer[2], _floatData[inMirfBuffer[2]]);
        else writeMirfAck(false);
        break;
      
      case USB_SET_BYTE: // 5- format: CMD, ID, dataID(byte), data(byte), SEP
        if (setByteData(inMirfBuffer[2], inMirfBuffer[3])) {
          onByteEvent(inMirfBuffer[2], inMirfBuffer[3]);
        }
        else writeMirfAck(false);
        break;
      
      case USB_SET_CHAR: // 6- format: CMD, ID, dataID(byte), data(char16), SEP
        ActoChar char16;

        for (int i=0; i < sizeof(char16); i++){//2
          char16.binary[i] = inMirfBuffer[(3+i)];
        }
        
        if (setCharData(inMirfBuffer[2], char16.character)) {
          onCharEvent(inMirfBuffer[2], char16.character);
        }
        else writeMirfAck(false);
        break;
      
      case USB_SET_INT:// 8- format: CMD, ID, dataID(byte), data(int), SEP
        ActoInt int32;
        
        for (int i=0; i < sizeof(int32); i++){//4
          int32.binary[i] = inMirfBuffer[(3+i)];
        }
        
        if (setIntData(inMirfBuffer[2], int32.integer)) {
          onIntEvent(inMirfBuffer[2], int32.integer);
        }
        else writeMirfAck(false);
        break;
      
      case USB_SET_FLOAT: // 8- format: CMD, ID, dataID(byte), data(int), SEP
        ActoFloat float32;
        
        for (int i=0; i < 4; i++){//4
          float32.binary[i] = inMirfBuffer[3+i];
        }
        
        if (setFloatData(inMirfBuffer[2], float32.floatingPoint)) {
          onFloatEvent(inMirfBuffer[2], float32.floatingPoint);
        }
        else writeMirfAck(false);
        break;
      
      
      case USB_MOTOR_MS:  // 9- format: CMD, ID, dataID(byte), data(byte), data(int), SEP        
        ActoInt int32_time;
        
        for (int i=0; i < sizeof(int32_time); i++){//4
          int32_time.binary[i] = inMirfBuffer[(4+i)];
        }

        driveDirectionSpeed(inMirfBuffer[2], inMirfBuffer[3], int32_time.integer);
        
        //writeSimpleMsg(USB_PING);
        break;
      
      default: 
        errorHandler(MIRFERROR);
    }
  }
}

// errorHandler blinks if error occurred
void errorHandler(int errorCode)
{
  if (STATUSLED >= 0) {  //if STATUSLED is available at all
    if (errorCode > 0)   // if new error
    {    
      errorTimestamp = millis() - ERRORINTERVAL;
      blinkNumber = errorCode;    
    }
    //check if time after last blink is past, if so change LED state
    if((millis() - errorTimestamp) >= ERRORINTERVAL) {
      if (blinkNumber > 0)
      {
        if (ledState == LOW) ledState = HIGH;
        else ledState = LOW;
        errorTimestamp = millis(); 
        blinkNumber--;  //count down number of blinks
      }
      else ledState = LOW;
      digitalWrite(STATUSLED,ledState);
    }
  }
}

boolean checkBattery() {
    //check if time after last blink is past, if so change LED state
    if((millis() - batteryTimestamp) >= BATTERYINTERVAL) {
      batteryTimestamp = millis();

      if (getBatteryVoltage() < BATTERYTHRESH)
      {
        errorHandler(BATTERYERROR);
        //stop motor
        initMotorUnit();
        
        //freeze completely and blink error
        while(true) {
          if (STATUSLED >= 0) {
            errorHandler(0);  
          }
          else {
            SPI.end();
            digitalWrite(13,HIGH);
            delay(200);
            digitalWrite(13,LOW);
            delay(200);
          }
        }
//        return false;
      }
  }
  return true;
}

float getBatteryVoltage() {
  int readVal = analogRead(BATTERYPIN);
  float currVolt = (readVal/RESISTANCEFACTOR)*REFERENCEVOLTAGE*4.0;

  //Serial.println(currVolt);
  return currVolt;
}

void writeMirfAck(boolean success) {
  if (success)
  {
    //writeSimpleMsg(USB_ACK);  //has to include ID, if used as handshake!!!! DOES NOT WORK NOW!
  }
  else
  {
    errorHandler(MIRFERROR);
    //writeSimpleMsg(USB_NAK);  //has to include ID, if used as handshake!
  }
}

//write simple Msg to ACTO_framework
void writeSimpleMsg(byte usb_msg) {
  //usb_msg, id, SEP
  outMirfBuffer[outMirfIndex] = usb_msg;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = mirfHomeAddress[4];
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = USB_SEP;
  sendMirfMsg();
}

//write BYTE Msg to ACTO_framework
void writeByteMsg(byte usb_msg, byte dataId, byte data) {
  //usb_msg, id, dataid, data(byte), SEP
  outMirfBuffer[outMirfIndex] = usb_msg;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = mirfHomeAddress[4];
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = dataId;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = data;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = USB_SEP;
  sendMirfMsg();
}

//write CHAR Msg to ACTO_framework
void writeCharMsg(byte usb_msg, byte dataId, char data) {
    //usb_msg, id, dataid, data(char), SEP
  outMirfBuffer[outMirfIndex] = usb_msg;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = mirfHomeAddress[4];
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = dataId;
  outMirfIndex++;

  ActoChar data16;
  data16.character = data; //2 RICHTIG  //TESTEN!!!
  //byte * b = (byte *) &data16.binary;
  
  for (int i=0; i < sizeof(data16); i++){//2
    outMirfBuffer[outMirfIndex] = data16.binary[i];
    outMirfIndex++;
  }
  
  outMirfBuffer[outMirfIndex] = USB_SEP;
  sendMirfMsg();
}
  
//write INT Msg to ACTO_framework
void writeIntMsg(byte usb_msg, byte dataId, int data) {
    //usb_msg, id, dataid, data(int), SEP
  outMirfBuffer[outMirfIndex] = usb_msg;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = mirfHomeAddress[4];
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = dataId;
  outMirfIndex++;

  ActoInt data32;
  data32.integer = (long)data;
  //long data32 = (long)data;   
  //byte * b = (byte *) &data32;  
  
  for (int i=0; i < sizeof(data32); i++){//4
    outMirfBuffer[outMirfIndex] = data32.binary[i];
    outMirfIndex++;
  }
  
  outMirfBuffer[outMirfIndex] = USB_SEP;
  sendMirfMsg();
}
  
//write FLOAT Msg to ACTO_framework
void writeFloatMsg(byte usb_msg, byte dataId, float data) {
  //usb_msg, id, dataid, data(float), SEP
  outMirfBuffer[outMirfIndex] = usb_msg;
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = mirfHomeAddress[4];
  outMirfIndex++;
  outMirfBuffer[outMirfIndex] = dataId;
  outMirfIndex++;

  ActoFloat data32;
  data32.floatingPoint = data;
  //byte * b = (byte *) &data;     
  
  for (int i=0; i < sizeof(data32); i++){//4
    outMirfBuffer[outMirfIndex] = data32.binary[i];
    outMirfIndex++;
  }
  
  outMirfBuffer[outMirfIndex] = USB_SEP;
  sendMirfMsg();
}

void sendMirfMsg() {
//possible collision, if trying incoming msg while sending, either handshake, or other method  
  if (outMirfIndex > 0) {
    Mirf.send((byte *)&outMirfBuffer);  //ZUVIEL, nur bis index!!
    //pause until ready
    while(Mirf.isSending()){
    }
    resetMirfBuffer();
  }
}

void resetMirfBuffer() {
  outMirfIndex = 0;
  memset(outMirfBuffer,0,mirfPayload);
}





//while(!Mirf.rxFifoEmpty()); //read as long as there is data received
//memcpy(outMirfBuffer, inMirfBuffer, mirfPayload);
