

//////////////////////////////////////////////////////////////////////////////////
// OS for ACTOframework serverBoard
// 
// project: ACTO
// author: Emanuel Vonach, Vienna University of Technology - IMS
//
// date: 25. Feb 2014
//
// version: v1.1
//
// OS for Arduino Diecimila/Uno operated via Android device
// allow communication from ACTO_framework via Nordic nRF24L01 chip to remote TUIOs running TUIO firmware
// using Mirf, a Arduino library for communication with nRF24L01
// http://playground.arduino.cc/InterfacingWithHardware/Nrf24L01
//
// see also ACTO_framework for protocol
//////////////////////////////////////////////////////////////////////////////////

#include "ACTOControl.h"

#include <Mirf.h>
#include <MirfHardwareSpiDriver.h>  //MMI
#include <MirfSpiDriver.h>  //MMI
#include <nRF24L01.h>

#include <SPI.h>

///////////////////////////////////////////////////////////////////////////////////////
// BASE SETTINGS
///////////////////////////////////////////////////////////////////////////////////////

//base id
const byte baseId = '1';

// for serial communication
const long baudRate = 115200;

//LED for error output
#define STATUSLED A0

//for Mirf communication
const byte mirfHomeAddress[5] = {'b','a','s','e', baseId};   //base
byte mirfAwayAddress[5] = {'t','u','i','o', 0};

// variables for errorHandler
unsigned long errorTimestamp = 0;
int blinkNumber = 0;
int ledState = LOW;

// for Mirf
byte outMirfBuffer[mirfPayload];
byte inMirfBuffer[mirfPayload];
int outMirfIndex = 0;

//for Serial communication
char incomingByte[1];
byte incomingCmd = 0;


void setup()   {
  
  //setup pins & stuff
  if (STATUSLED >= 0) pinMode(STATUSLED, OUTPUT);
  
  //open serial connection to Android device
  Serial.begin(baudRate);
  
  // MIRF SETUP  
  // ------------------------------------------------------------------------------------------------------------------------------// 
  // nRF24L01 |   ATmega8      | ATmega168      |    ARDUINO Diecimila (ATmega168)  |        Description                           //  
  // ------------------------------------------------------------------------------------------------------------------------------// 
  // SCK      |   PB5 (SCK)    |  PB5 (SCK)     |    digital pin 13                 |        SPI clock signal                      //  
  // MISO     |   PB4 (MISO)   |  PB4 (MISO)    |    digital pin 12                 |        SPI master in, slave out              //
  // MOSI     |   PB3 (MOSI)   |  PB3 (MOSI)    |    digital pin 11                 |        SPI master out, slave in              //
  // CSN      |   PB1          |  PB1           |    digital pin 9                  |        SPI Chip Select                       //
  // CE       |   PB0          |  PB0           |    digital pin 8                  |        Chip Enable Activates RX or TX mode   //
  // IRQ      |   PD2 (INT0)   |  PD6 (PCINT22) |    digital pin 6                  |        Interrupt pin. Active low             //
  // ------------------------------------------------------------------------------------------------------------------------------// 

  // setup pins / SPI
  Mirf.csnPin = 9;
  Mirf.cePin = 8;
  Mirf.spi = &MirfHardwareSpi;
  Mirf.init();

  // Configure Mirf parameters
  Mirf.setRADDR((byte*)mirfHomeAddress);  
  Mirf.payload = mirfPayload;
 
  // Write channel and payload config then power up reciver.
  Mirf.config();
}

void loop() {

  // if an error has occured previously, the LED changes state
  errorHandler(0);
  
  // listen to serial port  //change listening direction (Serial<->Mirf) after every message to avoid starving
  if (Serial.available() > 0)
  {
      // read first byte
//    Serial.readBytes(incomingByte,1);
//    incomingCmd = incomingByte[0];
    incomingCmd = (byte)Serial.read();
    
    switch (incomingCmd) {
      case USB_ACK: case USB_NAK:  // 2- format: CMD, SEP
        errorHandler(DEBUG);
        break;
        
      case USB_PING: case USB_SYNC:  // 3- format: CMD, ID, SEP
        addToMirfBuffer(incomingCmd, 3);
        break;
        
        //format: 4- CMD, ID, dataID(byte), SEP
      case USB_GET_DI_UNIT: case USB_GET_AN_UNIT: case USB_GET_BYTE: case USB_GET_CHAR: case USB_GET_INT: case USB_GET_FLOAT: 
        addToMirfBuffer(incomingCmd, 4);
        break;
        
      case USB_MOTOR: // 5- format: CMD, ID, direction(byte), speed(byte), SEP
      case USB_SET_DI_UNIT: case USB_SET_BYTE: // 5- format: CMD, ID, dataID(byte), data(byte), SEP
        addToMirfBuffer(incomingCmd, 5);
        break;
      
      case USB_SET_CHAR: // 6- format: CMD, ID, dataID(byte), data(char16), SEP
        addToMirfBuffer(incomingCmd, 6);
        break;
      
      case USB_SET_AN_UNIT: //only as answer!!, should be removed
      case USB_SET_PWM_UNIT: case USB_SET_INT: case USB_SET_FLOAT: // 8- format: CMD, ID, dataID(byte), data(int), SEP
        addToMirfBuffer(incomingCmd, 8);
        break;
        
       case USB_MOTOR_MS:  // 9- format: CMD, ID, dataID(byte), data(byte), data(int), SEP
        addToMirfBuffer(incomingCmd, 9);
        break;     
        
      default: 
        writeSerialAck(false);
    }
    sendMirfMsg();
  }

  if(Mirf.dataReady()){  //there should be no problem with crosstalk, since message is in payload as a whole
    //read message
    Mirf.getData((byte *)&inMirfBuffer);

    switch (inMirfBuffer[0]) {
//      case USB_ACK: case USB_NAK:  // 2- format: CMD, SEP  //DOES NOT WORK WITHOUT ID, PING is used instead
//        Serial.write(inMirfBuffer,2);
//        errorHandler(DEBUG);
//        break;
        
      case USB_PING: case USB_SYNC:  // 3- format: CMD, ID, SEP
        Serial.write(inMirfBuffer,3);
        break;
        
        //format: 4- CMD, ID, dataID(byte), SEP
      case USB_GET_DI_UNIT: case USB_GET_AN_UNIT: case USB_GET_BYTE: case USB_GET_CHAR: case USB_GET_INT: case USB_GET_FLOAT: 
        Serial.write(inMirfBuffer,4);
        break;
      
      case USB_MOTOR:
      case USB_SET_DI_UNIT: case USB_SET_BYTE: // 5- format: CMD, ID, dataID(byte), data(byte), SEP
        Serial.write(inMirfBuffer,5);
        break;
      
      case USB_SET_CHAR: // 6- format: CMD, ID, dataID(byte), data(char16), SEP
        Serial.write(inMirfBuffer,6);
        break;
      
      case USB_SET_PWM_UNIT: //should be removed
      case USB_SET_AN_UNIT: case USB_SET_INT: case USB_SET_FLOAT: // 8- format: CMD, ID, dataID(byte), data(int), SEP
        Serial.write(inMirfBuffer,8);
        break;
        
      case USB_MOTOR_MS:  // 9- format: CMD, ID, dataID(byte), data(byte), data(int), SEP
        Serial.write(inMirfBuffer,9);
        
      default: 
        writeSerialAck(false);
    }
    while (!(UCSR0A & _BV(TXC0)));  //wait Serial to finish writing
 }

}
////////////////////
// END loop
////////////////////


////////////////////
// errorHandler
//
// for error handling
// STATUSLED can blink to indicate an error or
// errorCode can be sent over Serial Port
////////////////////

void errorHandler(int errorCode)
{
  if (errorCode > 0)
  {    
    errorTimestamp = millis() - ERRORINTERVAL;
//    // Error messages over serial output could be specified here
//    if (errorCode == serialError || errorCode == serialErrorInvalidLED || errorCode == serialErrorInvalidFrame)
//    {
//      sendSerialACK(false);
//      Serial.write(serialError);
//    }
//    else //if(errorCode == I2CError)
//    {
      blinkNumber = errorCode;    
//    }
  }
  if((millis() - errorTimestamp) >= ERRORINTERVAL) {
    if (blinkNumber > 0)
    {
      if (ledState == LOW) ledState = HIGH;
      else ledState = LOW;
      errorTimestamp = millis(); 
      blinkNumber--;
    }
    else ledState = LOW;
    digitalWrite(STATUSLED,ledState);
  }
}


void writeSerialAck(boolean success) {
  if (success)
  {
    Serial.write(USB_ACK); 
  }
  else
  {
    errorHandler(SERIALERROR);
    Serial.write(USB_NAK);
  }
  Serial.write(USB_SEP);
}


boolean addToMirfBuffer (byte msgHeader, int msgSize) {
  
  outMirfBuffer[outMirfIndex] = msgHeader;
  outMirfIndex++;
  
  while (outMirfIndex < msgSize-1) {
    while(Serial.available() == 0);  //wait for rest of message
    outMirfBuffer[outMirfIndex] = Serial.read();
    outMirfIndex++;
  }
  while(Serial.available() == 0);  //wait for rest of message
  if (Serial.read() == USB_SEP)
  {
    outMirfBuffer[outMirfIndex] = USB_SEP;
    writeSerialAck(true);
    return true;
  }
  else
  {
    resetMirfBuffer();
    writeSerialAck(false);
  }
  return false;
}

void resetMirfBuffer() {
  outMirfIndex = 0;
  memset(outMirfBuffer,0,mirfPayload);
}

void sendMirfMsg() {
  if (outMirfIndex > 0) {
    mirfAwayAddress[4] = outMirfBuffer[1];  //set Tuio ID according to message
    Mirf.setTADDR((byte*)mirfAwayAddress);
    
    Mirf.send((byte *)&outMirfBuffer);

    //pause until ready
    while(Mirf.isSending()){
    }
    resetMirfBuffer();
  }
}


